using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace XELF.Tool {

	public class Pointer : DrawableGameComponent {
		Vector2 position;
		MouseState state;
		public Pointer(Game game)
			: base(game) {
		}
		public MouseState MouseState { get { return state; } }
		public override void Update(GameTime gameTime) {
			state = Mouse.GetState();
			position.X = state.X;
			position.Y = state.Y;
			base.Update(gameTime);
		}
		public Vector2 Position {
			get { return position; }
		}
	}

	public class Tools : DrawableGameComponent {
		SpriteFont font;
		readonly Pointer pointer;
		readonly KeyMap keys;

		public SpriteFont Font { get { return font; } }
		public Pointer Pointer { get { return pointer; } }
		public KeyMap Keys { get { return keys; } }
		public Tools(Game game)
			: base(game) {
			Game.Components.Add(pointer = new Pointer(game));
			Game.Components.Add(new MouseHUD(this));
			Game.Components.Add(keys = new KeyMap(game));
		}
		protected override void LoadContent() {
			font = Game.Content.Load<SpriteFont>("font/default");
			base.LoadContent();
		}
	}

	public class PointerHUD : DrawableGameComponent {
		SpriteBatch spriteBatch;
		readonly Canvas canvas;
		readonly Vector4[] image = new Vector4[1];

		public PointerHUD(Canvas canvas)
			: base(canvas.Game) {
			this.canvas = canvas;
			DrawOrder = 0x10000000;
		}

		protected override void LoadContent() {
			spriteBatch = new SpriteBatch(GraphicsDevice);
			base.LoadContent();
		}
		public override void Draw(GameTime gameTime) {
			Vector2 point = canvas.Pointed;
			Point location;
			location.X = (int)point.X;
			location.Y = (int)point.Y;

			if (canvas.Border.Contains(location)) {
				GraphicsDevice.Textures[0] = null;
				canvas.Texture.GetData(0, new Rectangle(location.X, location.Y, 1, 1), image, 0, 1);
				spriteBatch.Begin();
				spriteBatch.DrawString(canvas.Tools.Font,
					string.Format("Size=({0},{1})\n({2},{3}) : Color={4}",
					canvas.Texture.Width, canvas.Texture.Height,
					location.X, location.Y, image[0].ToString()),
					new Vector2(0x180, 0x200), Color.White);
				spriteBatch.End();
			}

			base.Draw(gameTime);
		}
	}

	public class Pencil : DrawableGameComponent {
		readonly Canvas canvas;
		readonly Vector4[] image = new Vector4[1] { new Vector4(1, 1, 1, 1) };

		public Pencil(Canvas canvas) : base(canvas.Game) { this.canvas = canvas; }

		public override void Update(GameTime gameTime) {
			if (!Game.IsActive)
				return;
			if (canvas.Tools.Pointer.MouseState.LeftButton == ButtonState.Pressed
				|| canvas.Tools.Keys.IsDown(Keys.Z)) {
				Vector2 point = canvas.Pointed;
				Point location;
				location.X = (int)point.X;
				location.Y = (int)point.Y;

				if (canvas.Border.Contains(location)) {
					canvas.Texture.SetData(0, new Rectangle(location.X, location.Y, 1, 1),
						image, 0, 1, SetDataOptions.None);
				}
			}
			if (canvas.Tools.Pointer.MouseState.RightButton == ButtonState.Pressed
				|| canvas.Tools.Keys.IsDown(Keys.X)) {
				Vector2 point = canvas.Pointed;
				Point location;
				location.X = (int)point.X;
				location.Y = (int)point.Y;

				if (canvas.Border.Contains(location)) {
					canvas.Texture.GetData(0, new Rectangle(location.X, location.Y, 1, 1),
						image, 0, 1);
				}
			}
		}
	}

	public class KeyMap : GameComponent {
		KeyboardState previous;
		KeyboardState current;

		public KeyMap(Game game)
			: base(game) {
		}

		public override void Update(GameTime gameTime) {
			previous = current;
			current = Keyboard.GetState();
			base.Update(gameTime);
		}
		public bool IsPress(Keys key) {
			return current.IsKeyDown(key) && previous.IsKeyUp(key);
		}
		public bool IsDown(Keys key) {
			return current.IsKeyDown(key);
		}
		public bool IsControlDown {
			get {
				return current.IsKeyDown(Keys.LeftControl) || current.IsKeyDown(Keys.RightControl);
			}
		}
		public bool IsPress(params Keys[] keys) {
			foreach (Keys key in keys) {
				if (!IsPress(key))
					return false;
			}
			return true;
		}
	}

	public class Canvas : DrawableGameComponent {
		readonly Tools tools;
		SpriteBatch spriteBatch;
		Texture2D background;
		Texture2D texture;
		float scale = 8.0f;
		Vector4[] image;
		Matrix transform = Matrix.Identity;
		Matrix inverseTransform = Matrix.Identity;
		string path = "untitled";
		readonly TextureFromFile textureFromFile = new TextureFromFile();

		public Canvas(Tools tools)
			: base(tools.Game) {
			this.tools = tools;

			Game.Components.Add(new PointerHUD(this));
			Game.Components.Add(new Pencil(this));

			UpdateTransform();
		}

		void UpdateTransform() {
			transform = Matrix.CreateScale(scale);
			inverseTransform = Matrix.Invert(transform);
		}
		public float Scale {
			get { return scale; }
			set {
				scale = value;
				UpdateTransform();
			}
		}
		public Vector2 Pointed {
			get {
				return Vector2.Transform(tools.Pointer.Position, inverseTransform);
			}
		}
		public Texture2D Texture {
			get { return texture; }
			set {
				texture = value;
				image = new Vector4[texture.Width * texture.Height];
				background = new Texture2D(
					GraphicsDevice, texture.Width, texture.Height, 1, TextureUsage.None, SurfaceFormat.Vector4);
				for (int y = 0; y < texture.Height; y++) {
					for (int x = 0; x < texture.Width; x++) {
						image[x + y * texture.Width] = new Vector4(1, 1, 1, 1);
					}
				}
				background.SetData(image);
			}
		}
		public Tools Tools { get { return tools; } }
		public Rectangle Border { get { return new Rectangle(0, 0, texture.Width, texture.Height); } }
		protected override void LoadContent() {
			spriteBatch = new SpriteBatch(GraphicsDevice);
			Texture = new Texture2D(GraphicsDevice, 64, 64, 1, TextureUsage.None, SurfaceFormat.Vector4);

			base.LoadContent();
		}

		void Load() {
			System.Windows.Forms.OpenFileDialog dialog = new System.Windows.Forms.OpenFileDialog();
			dialog.Filter
				= "Portable Network Graphics|*.png|"
				+ "Bitmap|*.bmp|"
				+ "Joint Photographic Experts Group|*.jpg;*.jpeg|"
				+ "Enhanced Meta File|*.emf|"
				+ "Graphics Interchange Format|*.gif|"
				+ "Exchangeable Image Format|*.exif|"
				+ "Windows Icon Image|*.icon|"
				+ "Tagged Image File Format|*.tiff|"
				+ "Windows Meta File|*.wmf|"
				+ "Supported Files|*.png;*.bmp;*.jpg;*.jpeg;*.emf;*.gif;*.exif;*.icon;*.tiff;*.wmf";
			if (dialog.ShowDialog() == System.Windows.Forms.DialogResult.OK) {
				Texture = textureFromFile.Load(GraphicsDevice, dialog.FileName);
				path = dialog.FileName;
			}
		}
		void Save() {
			System.Windows.Forms.SaveFileDialog dialog = new System.Windows.Forms.SaveFileDialog();
			dialog.FileName = path;
			dialog.Filter
				= "Portable Network Graphics|*.png|"
				+ "Joint Photographic Experts Group|*.jpg;*.jpeg|"
				+ "DirectDraw Surface|*.dds|"
				+ "Portable Pixel Map|*.ppm|"
				+ "(*.pfm)|*.pfm|"
				+ "Truevision Advanced Raster Graphics Adapter|*.tga|"
				+ "(*.hdr)|*.hdr|"
				+ "Bitmap|*.bmp|"
				+ "Device Independent Bitmap|*.dib";
			if (dialog.ShowDialog() == System.Windows.Forms.DialogResult.OK) {
				int index = dialog.FileName.LastIndexOf('.');
				string extension = dialog.FileName.Substring(index + 1);
				ImageFileFormat format;
				if (TextureFromFile.ImageFileFormatFromExtension(extension, out format)) {
					texture.Save(dialog.FileName, format);
					path = dialog.FileName;
				}
			}
		}

		public override void Update(GameTime gameTime) {
			if (tools.Keys.IsPress(Keys.OemPlus)) {
				Scale *= 2;
			}
			if (tools.Keys.IsPress(Keys.OemMinus)) {
				Scale *= 0.5f;
			}
			if (tools.Keys.IsPress(Keys.Left)) {
				Mouse.SetPosition(tools.Pointer.MouseState.X - 1, tools.Pointer.MouseState.Y);
			}
			if (tools.Keys.IsPress(Keys.Right)) {
				Mouse.SetPosition(tools.Pointer.MouseState.X + 1, tools.Pointer.MouseState.Y);
			}
			if (tools.Keys.IsPress(Keys.Up)) {
				Mouse.SetPosition(tools.Pointer.MouseState.X, tools.Pointer.MouseState.Y - 1);
			}
			if (tools.Keys.IsPress(Keys.Down)) {
				Mouse.SetPosition(tools.Pointer.MouseState.X, tools.Pointer.MouseState.Y + 1);
			}
			if (tools.Keys.IsControlDown && tools.Keys.IsPress(Keys.O)) {
				Load();
			}
			if (tools.Keys.IsControlDown && tools.Keys.IsPress(Keys.S)) {
				Save();
			}
			base.Update(gameTime);
		}
		public override void Draw(GameTime gameTime) {
			spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Immediate, SaveStateMode.SaveState);
			GraphicsDevice.SamplerStates[0].MagFilter = TextureFilter.Point;

			Vector2 center = new Vector2(texture.Width, texture.Height) * (scale * 0.5f);
			Vector2 origin = new Vector2(texture.Width, texture.Height) * 0.5f;
			spriteBatch.Draw(background, center, null, Color.Black, 0, origin, scale, SpriteEffects.None, 0);
			spriteBatch.Draw(texture, center, null, Color.White, 0, origin, scale, SpriteEffects.None, 0);
			spriteBatch.End();
			base.Draw(gameTime);
		}
	}

	public class TextureFromFile {
		readonly byte[] buffer = new byte[4];

		public static bool ImageFileFormatFromExtension(string extension, out ImageFileFormat format) {
			switch (extension.ToUpper()) {
			default:
				format = ImageFileFormat.Png;
				return false;
			case "PNG":
				format = ImageFileFormat.Png;
				break;
			case "JPG":
			case "JPEG":
				format = ImageFileFormat.Jpg;
				break;
			case "BMP":
				format = ImageFileFormat.Bmp;
				break;
			case "DDS":
				format = ImageFileFormat.Dds;
				break;
			case "DIB":
				format = ImageFileFormat.Dib;
				break;
			case "TGA":
				format = ImageFileFormat.Tga;
				break;
			case "PPM":
				format = ImageFileFormat.Ppm;
				break;
			case "PFM":
				format = ImageFileFormat.Pfm;
				break;
			case "HDR":
				format = ImageFileFormat.Hdr;
				break;
			}
			return true;
		}

		public Texture2D Load(GraphicsDevice device, string fileName) {
			using (System.Drawing.Bitmap bitmap
				= (System.Drawing.Bitmap)System.Drawing.Bitmap.FromFile(fileName)) {
				System.Drawing.Imaging.BitmapData data = bitmap.LockBits(
					new System.Drawing.Rectangle(0, 0, bitmap.Width, bitmap.Height),
					System.Drawing.Imaging.ImageLockMode.ReadOnly,
					 System.Drawing.Imaging.PixelFormat.Format32bppArgb);

				Texture2D texture = new Texture2D(
					device, bitmap.Width, bitmap.Height, 1, TextureUsage.None, SurfaceFormat.Vector4);
				Vector4[] image = new Vector4[bitmap.Width * bitmap.Height];

				long s = data.Scan0.ToInt64();
				int i = 0;
				for (int y = 0; y < bitmap.Height; y++) {
					long r = s + y * data.Stride;
					for (int x = 0; x < bitmap.Width; x++) {
						Marshal.Copy(new IntPtr(r + x * 4), buffer, 0, 4);
						image[i].X = buffer[2] * (1.0f / byte.MaxValue);
						image[i].Y = buffer[1] * (1.0f / byte.MaxValue);
						image[i].Z = buffer[0] * (1.0f / byte.MaxValue);
						image[i].W = buffer[3] * (1.0f / byte.MaxValue);
						i++;
					}
				}
				bitmap.UnlockBits(data);

				texture.SetData(image);
				return texture;
			}
		}
	}

	public class MouseHUD : DrawableGameComponent {
		SpriteBatch spriteBatch;
		SpriteFont font;
		readonly Tools tools;

		public MouseHUD(Tools tools)
			: base(tools.Game) {
			this.tools = tools;
			DrawOrder = 0x10000000;
		}

		protected override void LoadContent() {
			spriteBatch = new SpriteBatch(GraphicsDevice);
			font = Game.Content.Load<SpriteFont>("font/default");
		}

		public override void Draw(GameTime gameTime) {
			MouseState mouseState = tools.Pointer.MouseState;
			bool left = mouseState.LeftButton == ButtonState.Pressed;
			bool right = mouseState.RightButton == ButtonState.Pressed;
			bool middle = mouseState.MiddleButton == ButtonState.Pressed;
			bool xbutton1 = mouseState.XButton1 == ButtonState.Pressed;
			bool xbutton2 = mouseState.XButton2 == ButtonState.Pressed;
			spriteBatch.Begin();
			spriteBatch.DrawString(font,
				string.Format("({0},{1} {2} {3} {4} {5} {6})", mouseState.X, mouseState.Y,
				left ? "L " : "- ",
				middle ? "M " : "- ",
				right ? "R " : "- ",
				xbutton1 ? "X1" : "- ",
				xbutton2 ? "X2" : "- "
				), Vector2.Zero, Color.White);
			spriteBatch.End();

			base.Draw(gameTime);
		}
	}

	public class ToolGame : Microsoft.Xna.Framework.Game {
		readonly GraphicsDeviceManager graphics;
		readonly Tools tools;
		readonly Canvas canvas;

		public ToolGame() {
			graphics = new GraphicsDeviceManager(this);
			graphics.SynchronizeWithVerticalRetrace = false;
			Content.RootDirectory = "Content";
			IsMouseVisible = true;
			TargetElapsedTime = TimeSpan.FromMilliseconds(1);
			IsFixedTimeStep = false;
			Window.AllowUserResizing = true;
			Components.Add(tools = new Tools(this));
			Components.Add(canvas = new Canvas(tools));
		}

		protected override void Draw(GameTime gameTime) {
			graphics.GraphicsDevice.Clear(Color.CornflowerBlue);
			base.Draw(gameTime);
		}
	}
}
